
from array import array

from ._network import Network

__all__ = ["initialise_infections",
           "initialise_play_infections"]


def initialise_infections(network: Network):
    """Initialise the data structure used to store the infections"""

    params = network.params

    if params is None:
        return None

    disease = params.disease_params

    n = disease.N_INF_CLASSES()

    infections = []

    nlinks = network.nlinks + 1

    int_t = "i"
    null_int = nlinks * [0]

    for _ in range(0, n):
        infections.append( array(int_t, null_int) )

    return infections


def initialise_play_infections(network: Network):
    """Initialise the space used to store the play infections"""

    params = network.params

    if params is None:
        return None

    disease = params.disease_params

    n = disease.N_INF_CLASSES()

    infections = []

    nnodes = network.nnodes + 1

    int_t = "i"
    null_int = nnodes * [0]

    for _ in range(0, n):
        infections.append( array(int_t, null_int) )

    return infections




import os

__all__ = ["open_files"]


def open_files(output_dir: str):
    """Opens all of the output files written to during the simulation,
       opening them all in the directory 'output_dir'

       This returns the file handles of all open files
    """
    files = []

    files.append( open(os.path.join(output_dir, "WorkInfections.dat"), "w") )
    files.append( open(os.path.join(output_dir, "NumberWardsInfected.dat"),
                       "w") )
    files.append( open(os.path.join(output_dir, "MeanXY.dat"), "w") )
    files.append( open(os.path.join(output_dir, "PlayInfections.dat"), "w") )
    files.append( open(os.path.join(output_dir, "TotalInfections.dat"), "w") )
    files.append( open(os.path.join(output_dir, "VarXY.dat"), "w") )
    files.append( open(os.path.join(output_dir, "Dispersal.dat"), "w") )

    return files




cimport cython
from libc.stdint cimport uintptr_t

cdef extern from "gsl/gsl_rng.h":
    ctypedef struct gsl_rng_type
    ctypedef struct gsl_rng
    cdef gsl_rng_type *gsl_rng_default

    gsl_rng *gsl_rng_alloc ( gsl_rng_type * T) nogil
    void gsl_rng_free (gsl_rng * r) nogil

    void gsl_rng_free (gsl_rng * r) nogil

    void gsl_rng_set ( gsl_rng * r, unsigned long int seed) nogil


cdef extern from "gsl/gsl_randist.h":
    unsigned int gsl_ran_binomial ( gsl_rng * r, double p, unsigned int n) nogil


cdef inline uintptr_t _construct_gsl_rng() nogil:
    """Construct and return a pointer to a new gsl random number
       generator. This will return a uintptr_t that can be handled
       in Python. You must destroy this random number generator
       when you don't want it anymore, using the _delete_gsl_rng
       function
    """
    cdef gsl_rng_type * T
    cdef gsl_rng* r
    T = gsl_rng_default
    r = gsl_rng_alloc(T)

    cdef uintptr_t p = <uintptr_t>r

    return p


cdef inline void _delete_gsl_rng(uintptr_t rng) nogil:
    """Delete the passed gsl random number generator.
       WARNING: It is not safe to use this pointer to
       generate any more random numbers!
    """
    cdef gsl_rng* r = _get_gsl_ptr(rng)
    gsl_rng_free(r)


cdef inline gsl_rng* _get_gsl_ptr(uintptr_t rng) nogil:
    """Get the underlying gsl_rng pointer from the uintptr_t"""
    cdef gsl_rng* r = <gsl_rng*>rng

    return r


cdef inline uintptr_t _seed_ran_binomial(uintptr_t rng,
                                         unsigned long seed) nogil:
    """Seed the passed random number generator"""
    cdef gsl_rng* r = _get_gsl_ptr(rng)
    gsl_rng_set(r, seed)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _ran_binomial(gsl_rng *rng, double p, int n) nogil:
    """Generate a random number from the binomial distribution
       described by p and n
    """
    return gsl_ran_binomial(rng, p, n)



"""
Python version of the MetaWards program

Details of authorship / license etc to be written here

"""

from ._parameters import *
from ._disease import *
from ._inputfiles import *
from ._node import *
from ._link import *
from ._population import *
from ._network import *
from ._profiler import *

# All of the functions are now hidden in the 'utils' submodule
from . import _utils as utils

# import the pyx cython-compiled modules
from ._nodes import *
from ._links import *

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions




from ._network import Network
from ._parameters import Parameters

__all__ = ["infect_additional_seeds", "load_additional_seeds",
           "seed_infection_at_node", "seed_all_wards"]


def infect_additional_seeds(network: Network, params: Parameters,
                            infections, play_infections,
                            additional_seeds, timestep: int):
    """Cause more infection from additional infection seeds"""
    wards = network.nodes

    for seed in additional_seeds:
        if seed[0] == timestep:
            if wards.play_suscept[seed[1]] < seed[2]:
                print(f"Not enough susceptibles in ward for seeding")
            else:
                wards.play_suscept[seed[1]] -= seed[2]
                #print(f"seeding play_infections[0][{seed[1]}] += {seed[2]}")
                play_infections[0][seed[1]] += seed[2]


def load_additional_seeds(filename: str):
    """Load additional seeds from the passed filename. This returns
       the added seeds
    """
    print(f"Loading additional seeds from {filename}...")

    with open(filename, "r") as FILE:
        line = FILE.readline()
        seeds = []

        while line:
            words = line.split()

            # yes, this is really the order of the seeds - "t num loc"
            # is in the file as "t loc num"
            seeds.append( (int(words[0]), int(words[2]), int(words[1])) )
            print(seeds[-1])
            line = FILE.readline()

    return seeds


def seed_infection_at_node(network: Network, params: Parameters,
                           seed: int, infections, play_infections):
    """Seed the infection at a specific ward"""
    wards = network.nodes
    links = network.to_links

    j = 0

    while (links.ito[j] != seed) or (links.ifrom[j] != seed):
        j += 1

    #print(f"j {j} link from {links.ifrom[j]} to {links.ito[j]}")

    if links.suscept[j] < params.initial_inf:
        wards.play_suscept[seed] -= params.initial_inf
        #print(f"seed at play_infections[0][{seed}] += {params.initial_inf}")
        play_infections[0][seed] += params.initial_inf

    infections[0][j] = params.initial_inf
    links.suscept[j] -= params.initial_inf


def seed_all_wards(network: Network, play_infections,
                   expected: int, population: int):
    """Seed the wards with an initial set of infections, assuming
       an 'expected' number of infected people out of a population
       of 'population'
    """
    wards = network.nodes

    frac = float(expected) / float(population)

    for i in range(0, network.nnodes+1):  # 1-index but also count at 0?
        temp = wards.denominator_n[i] + wards.denominator_p[i]
        to_seed = int(frac*temp + 0.5)
        wards.play_suscept[i] -= to_seed
        play_infections[0][i] += to_seed




from dataclasses import dataclass
from typing import List
from copy import deepcopy
import pathlib
import os

from ._inputfiles import InputFiles
from ._disease import Disease

__all__ = ["Parameters", "get_repository_version"]

_default_parameters_path = os.path.join(pathlib.Path.home(),
                                        "GitHub", "MetaWardsData")

_default_folder_name = "parameters"


_repositories = {}

def get_repository_version(repository):
    """Read and return the Git version of the passed repository"""
    global _repositories

    if repository in _repositories:
        return _repositories[repository]

    filename = os.path.join(repository, "version.txt")

    try:
        with open(filename) as FILE:
            version = FILE.readline().strip()
            _repositories[repository] = version
            return version
    except Exception:
        print(f"Could not find the repository version info in {filename}."
              f"Please make sure that you have run './version' in that "
              f"repository to generate the version info.")
        _repositories[repository] = "unknown"
        return _repositories[repository]


@dataclass
class Parameters:
    input_files: InputFiles = None
    uv_filename: str = None
    disease_params: Disease = None

    length_day: float = 0.7
    plength_day: float = 0.5
    initial_inf: int = 5

    static_play_at_home: float = 0.0
    dyn_play_at_home: float = 0.0

    data_dist_cutoff: float = 10000000.0
    dyn_dist_cutoff: float = 10000000.0

    play_to_work: float = 0.0
    work_to_play: float = 0.0

    local_vaccination_thresh: int = 4
    global_detection_thresh: int = 4
    daily_ward_vaccination_capacity: int = 5
    neighbour_weight_threshold: float = 0.0

    daily_imports: float = 0.0 # proportion of daily imports
    UV: float = 0.0

    _name: str = None
    _version: str = None
    _authors: str = None
    _contacts: str = None
    _references: str = None
    _filename: str = None
    _repository: str = None
    _repository_version: str = None

    def __str__(self):
        return f"Parameters {self._name}\n" \
               f"loaded from {self._filename}\n" \
               f"version: {self._version}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_version: {self._repository_version}\n\n" \
               f"length_day = {self.length_day}\n" \
               f"plength_day = {self.plength_day}\n" \
               f"initial_inf = {self.initial_inf}\n" \
               f"static_play_at_home = {self.static_play_at_home}\n" \
               f"dyn_play_at_home = {self.dyn_play_at_home}\n" \
               f"data_dist_cutoff = {self.data_dist_cutoff}\n" \
               f"dyn_dist_cutoff = {self.dyn_dist_cutoff}\n" \
               f"play_to_work = {self.play_to_work}\n" \
               f"work_to_play = {self.work_to_play}\n" \
               f"local_vaccination_thresh = {self.local_vaccination_thresh}\n" \
               f"global_detection_thresh = {self.global_detection_thresh}\n" \
               f"daily_ward_vaccination_capacity = {self.daily_ward_vaccination_capacity}\n" \
               f"neighbour_weight_threshold = {self.neighbour_weight_threshold}\n" \
               f"daily_imports = {self.daily_imports}\n" \
               f"UV = {self.UV}\n\n"

    @staticmethod
    def load(parameters: str = "march29",
             repository: str = None,
             folder: str=_default_folder_name,
             filename: str = None):
        """ This will return a Parameters object containing all of the
            parameters loaded from the parameters found in file
            f"{repository}/{folder}/{parameters}.json"

            By default this will load the march29 parameters from
            $HOME/GitHub/model_data/2011Data/parameters/march29.json

            Alternatively, you can provide the exact path to the
            filename via the 'filename' argument
        """
        repository_version = None

        if filename is None:
            if repository is None:
                repository = os.getenv("METAWARDSDATA")
                if repository is None:
                    repository = _default_parameters_path

            repository_version = get_repository_version(repository)
            filename = os.path.join(repository, folder, f"{parameters}.json")

        json_file = filename

        try:
            with open(json_file, "r") as FILE:
                import json
                data = json.load(FILE)

        except Exception as e:
            print(f"Could not find the parameters file {json_file}")
            print(f"Either it does not exist of was corrupted.")
            print(f"Error was {e.__class__} {e}")
            print(f"To download the parameters type the command:")
            print(f"  git clone https://github.com/chryswoods/MetaWardsData")
            print(f"and then re-run this function passing in the full")
            print(f"path to where you downloaded this directory")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")


        par = Parameters(length_day=data["length_day"],
                         plength_day=data["plength_day"],
                         initial_inf=data["initial_inf"],
                         static_play_at_home=data["static_play_at_home"],
                         dyn_play_at_home=data["dyn_play_at_home"],
                         data_dist_cutoff=data["data_dist_cutoff"],
                         dyn_dist_cutoff=data["dyn_dist_cutoff"],
                         play_to_work=data["play_to_work"],
                         work_to_play=data["work_to_play"],
                         local_vaccination_thresh=data["local_vaccination_threshold"],
                         global_detection_thresh=data["global_detection_threshold"],
                         daily_ward_vaccination_capacity=data["daily_ward_vaccination_capacity"],
                         neighbour_weight_threshold=data["neighbour_weight_threshold"],
                         daily_imports=data["daily_imports"],
                         UV=data["UV"],
                         _name=data["name"],
                         _authors=data["author(s)"],
                         _version=data["version"],
                         _contacts=data["contact(s)"],
                         _references=data["reference(s)"],
                         _filename=json_file,
                         _repository=repository,
                         _repository_version=repository_version
                         )

        print("Using parameters:")
        print(par)

        return par

    def set_input_files(self, input_files: InputFiles):
        """Set the input files that are used to initialise the
           simulation
        """
        if isinstance(input_files, str):
            input_files = InputFiles.load(input_files,
                                          repository=self._repository)

        print("Using input files:")
        print(input_files)

        self.input_files = deepcopy(input_files)

    def set_disease(self, disease: Disease):
        """"Set the disease that will be modelled"""
        if isinstance(disease, str):
            disease = Disease.load(disease,
                                   repository=self._repository)

        print("Using disease")
        print(disease)

        self.disease_params = deepcopy(disease)

    def read_file(self, filename: str, line_number: int):
        """Read in extra parameters from the specified line number
           of the specified file
        """
        print(f"Reading in parameters from line {line_number} of {filename}")

        i = 0
        with open(filename, "r") as FILE:
            line = FILE.readline()

            if i == line_number:
                words = line.split(",")

                if len(words) != 5:
                    raise ValueError(
                        f"Corrupted input file. Expecting 5 values. "
                        f"Received {line}")

                vals = []

                try:
                    for word in words:
                        vals.append(float(word))
                except Exception:
                    raise ValueError(
                            f"Corrupted input file. Expected 5 numbers. "
                            f"Received {line}")

                self.disease_params.beta[2] = vals[0]
                self.disease_params.beta[3] = vals[1]
                self.disease_params.progress[1] = vals[2]
                self.disease_params.progress[2] = vals[3]
                self.disease_params.progress[3] = vals[4]

                print(f"Updated beta = {self.disease_params.beta}")
                print(f"Updated progress = {self.disease_params.progress}")

                return
            else:
                i += 1

        # get here if we can't find this line in the file
        raise ValueError(f"Cannot read parameters from line {line_number} "
                         f"as the file contains just {i} lines")




from dataclasses import dataclass
from typing import List

from ._parameters import Parameters
from ._nodes import Nodes
from ._links import Links
from ._population import Population
from ._ran_binomial import seed_ran_binomial

__all__ = ["Network"]


@dataclass
class Network:
    """This class represents a network of wards. The network comprises
       nodes (representing wards), connected with links which represent
       work (predictable) links. There are also additional links for
       play (unpredictable/random) and weekend
    """

    """The list of nodes (wards) in the network"""
    nodes: Nodes = None
    """The links between nodes (work)"""
    to_links: Links = None
    """The links between nodes (play)"""
    play: Links = None
    """The links between nodes (weekend)"""
    weekend: Links = None

    """The number of nodes in the network"""
    nnodes: int = 0
    """The number of links in the network"""
    nlinks: int = 0
    """The number of play links in the network"""
    plinks: int = 0

    """The maximum allowable number of nodes in the network"""
    max_nodes: int = 10050
    """The maximum allowable number of links in the network"""
    max_links: int = 2414000

    """To seed provides additional seeding information"""
    to_seed: List[int] = None

    params: Parameters = None  # The parameters used to generate this network

    @staticmethod
    def build(params: Parameters,
              calculate_distances: bool=True,
              build_function=None,
              distance_function=None,
              max_nodes:int = 10050,
              max_links:int = 2414000,
              profile:bool = True):
        """Builds and returns a new Network that is described by the
           passed parameters. If 'calculate_distances' is True, then
           this will also read in the ward positions and add
           the distances between the links.

           Optionally you can supply your own function to build the network,
           by supplying 'build_function'. By default, this is
           metawards.utils.build_wards_network.

           Optionally you can supply your own function to read and
           calculate the distances by supplying 'build_function'.
           By default this is metawards.add_wards_network_distance

           The network is built in allocated memory, so you need to specify
           the maximum possible number of nodes and links. The memory buffers
           will be shrunk back after building.
        """
        if profile:
            from metawards import Profiler
            p = Profiler()
        else:
            from metawards import NullProfiler
            p = NullProfiler()

        p = p.start("Network.build")

        if build_function is None:
            from ._utils import build_wards_network
            build_function = build_wards_network

        p = p.start("build_function")
        network = build_function(params=params,
                                 profiler=p,
                                 max_nodes=max_nodes,
                                 max_links=max_links)
        p = p.stop()

        if calculate_distances:
            p = p.start("add_distances")
            network.add_distances(distance_function=distance_function)
            p = p.stop()

        if params.input_files.seed:
            from ._utils import read_done_file
            p = p.start("read_done_file")
            to_seed = read_done_file(params.input_files.seed)
            nseeds = len(to_seed)

            print(to_seed)
            print(f"Number of seeds equals {nseeds}")
            network.to_seed = to_seed
            p = p.stop()

        # By default, we initialise the network ready for a run,
        # namely make sure everything is reset and the population
        # is at work
        print("Reset everything...")
        p = p.start("reset_everything")
        network.reset_everything()
        p = p.stop()

        print("Rescale play matrix...")
        p = p.start("rescale_play_matrix")
        network.rescale_play_matrix()
        p = p.stop()

        print("Move population from play to work...")
        p = p.start("move_from_play_to_work")
        network.move_from_play_to_work()
        p = p.stop()

        if not p.is_null():
            p = p.stop()
            print(p)

        return network

    def add_distances(self, distance_function=None):
        """Read in the positions of all of the nodes (wards) and calculate
           the distances of the links.

           Optionally you can specify the function to use to
           read the positions and calculate the distances.
           By default this is mw.utils.add_wards_network_distance
        """

        if distance_function is None:
            from ._utils import add_wards_network_distance
            distance_function = add_wards_network_distance

        distance_function(self)

        # now need to update the dynamic distance cutoff based on the
        # maximum distance between nodes
        print("Get min/max distances...")
        (_mindist, maxdist) = self.get_min_max_distances()

        self.params.dyn_dist_cutoff = maxdist + 1

    def initialise_infections(self):
        """Initialise and return the space that will be used
           to track infections
        """
        from ._utils import initialise_infections
        return initialise_infections(self)

    def initialise_play_infections(self):
        """Initialise and return the space that will be used
           to track play infections
        """
        from ._utils import initialise_play_infections
        return initialise_play_infections(self)

    def get_min_max_distances(self):
        """Calculate and return the minimum and maximum distances
           between nodes in the network
        """
        try:
            return self._min_max_distances
        except Exception:
            pass

        from ._utils import get_min_max_distances
        self._min_max_distances = get_min_max_distances(self)

        return self._min_max_distances

    def reset_everything(self):
        """Resets the network ready for a new run of the model"""
        from ._utils import reset_everything
        reset_everything(self)

    def rescale_play_matrix(self):
        """Rescale the play matrix"""
        from ._utils import rescale_play_matrix
        rescale_play_matrix(self)

    def move_from_play_to_work(self):
        """Move the population from play to work"""
        from ._utils import move_population_from_play_to_work
        move_population_from_play_to_work(self)

    def run(self, population: Population,
                  seed: int = None,
                  output_dir: str = "tmp",
                  nsteps: int = None,
                  profile: bool = True,
                  s: int = None):
        """Run the model simulation for the passed population.
           The random number seed is given in 'seed'. If this
           is None, then a random seed is used.

           All output files are written to 'output_dir'

           The simulation will continue until the infection has
           died out or until 'nsteps' has passed (keep as 'None'
           to prevent exiting early).

           s is used to select the 'to_seed' entry to seed
           the nodes
        """
        # Create the random number generator
        rng = seed_ran_binomial(seed=seed)

        # Create space to hold the results of the simulation
        print("Initialise infections...")
        infections = self.initialise_infections()

        print("Initialise play infections...")
        play_infections = self.initialise_play_infections()

        from ._utils import run_model
        population = run_model(network=self,
                               population=population.initial,
                               infections=infections,
                               play_infections=play_infections,
                               rng=rng, s=s, output_dir=output_dir,
                               nsteps=nsteps, profile=profile)

        # do we want to save infections and play_infections for inspection?

        return population
